import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.*;
import processing.opengl.*;
import java.nio.*;
import java.awt.event.*;

float vtx = 0, vty = 0;
float qx = 0, qy = 0, qs = 1, qt = 1;
float tx = 0, ty = 0;
float vsc = 1;
float sc = 1;

int noo = 0;
int not = 0;

int realm = 1;

FloatBuffer vbuf;

int[][] trace;


int window = 128;
int start = 0;

int frameNo = 0;
boolean auto = true;
boolean invert = true;

float clamp(float v, float lo, float hi) {
  return min(max(v,lo),hi);
}

void setup() {
  size(800,600, OPENGL);
  frameRate(60);

  try {
    frame.addMouseWheelListener(new MouseWheelInput());
  } 
  catch (NullPointerException npe) {
    println("couldn't grab the mouse wheel"); 
  }


  setup_occupancy();
  setup_trace();
}


void draw() {
  realm = min(width,height);
  if(invert) {
    background(0);
  } else {
    background(255);
  }

  PGraphicsOpenGL pgl = (PGraphicsOpenGL) g;  // g may change
  GL gl = pgl.beginGL();  // always use the GL object returned by beginGL

    gl.glPushMatrix();
  gl.glTranslatef(width/2,height/2,0);

  gl.glDisable(GL.GL_DEPTH_TEST);

  float lim = 20;
  gl.glRotatef(clamp(vty*sc*360,-lim,lim),1,0,0);
  gl.glRotatef(clamp(-vtx*sc*360,-lim,lim),0,1,0);
  gl.glScalef(realm,realm,realm);
  gl.glScalef(sc,sc,sc);
  gl.glTranslatef(tx,ty,0);

  // animation state
  float vk = 0.85;
  tx += vtx; 
  vtx *= vk;
  ty += vty; 
  vty *= vk;
  sc += vsc; 
  vsc *= vk*vk;

  sc = clamp(sc,1,1<<16);

  frameNo++;
  if(frameNo%1 == 0) {
    start = (start+1)%not;
  }

  if(auto) {
    float kt = 0.01;
    int q = (start+window)%not;
    vtx += kt*( -qx/qt -tx);
    vty += kt*( -qy/qt -ty);

    float ks = 0.01;
    vsc += ks*( qs/qt -sc);
    vsc -= 10*max(abs(vtx),abs(vty))*sc;
  }

  qx = 0;
  qy = 0;
  qs = 0;
  qt = 0;

  gl.glBegin(GL.GL_QUADS);
  for (int i=0; i<window; i++) {
    int idx = (i+start)%not;
    float pos = (float)i/window;
    int[] tile = trace[idx];
    float sz = 1.0/(1<<tile[2]);
    float hz = (1<<tile[2])/2.0;

    float qk = pow(pos,6);
    qx += qk*(tile[0]-hz+0.0)*sz;
    qy += qk*(tile[1]-hz+0.0)*sz;
    qs += qk*(1<<tile[2])/16.0;
    qt += qk;

    float chan = tile[2]/20.0;
    gl.glColor4f(pow(chan,3),1-pow(chan,3)-pow(1-chan,3),pow(1-chan,3),pow(pos,2));

    gl.glVertex3f((tile[0]-hz+0.0)*sz,(tile[1]-hz+0.0)*sz,0);
    gl.glVertex3f((tile[0]-hz+0.0)*sz,(tile[1]-hz+1.0)*sz,0);
    gl.glVertex3f((tile[0]-hz+1.0)*sz,(tile[1]-hz+1.0)*sz,0);
    gl.glVertex3f((tile[0]-hz+1.0)*sz,(tile[1]-hz+0.0)*sz,0);

  }
  gl.glEnd();

  gl.glLineWidth(1);
  if(invert) {
    gl.glColor4f(1,1,1,0.125);
  } else {
    gl.glColor4f(0,0,0,0.125);
  }
  gl.glDrawArrays(GL.GL_LINES, 0, noo*8);

  gl.glPopMatrix();

  pgl.endGL();

}

void keyPressed() {
  float d = (1.0/8.0)/sc;

  if(keyCode == ' ') { 
    auto = !auto;
  } else if (keyCode == 'I') {
    invert = !invert; 
  } else if (keyCode == KeyEvent.VK_PAGE_UP) {
    sc *= 2;
  } 
  else if (keyCode == KeyEvent.VK_PAGE_DOWN) {
    sc /= 2; 
  } 
  else if (keyCode == UP) {
    ty += d; 
  } 
  else if (keyCode == DOWN) {
    ty -= d; 
  } 
  else if (keyCode == LEFT) {
    tx += d; 
  } 
  else if (keyCode == RIGHT) {
    tx -= d; 
  }
}

void mouseDragged() {
  float dx = (mouseX - pmouseX) / (sc*realm);
  float dy = (mouseY - pmouseY) / (sc*realm);
  tx += dx;
  ty += dy;
}

class MouseWheelInput implements MouseWheelListener{  

  void mouseWheelMoved(MouseWheelEvent e) {  
    int step=e.getWheelRotation();  

    if(sc <= 1 && step > 0) { 
      return;
    }  
    if(sc >= 1<<16 && step < 0) { 
      return;
    }

    tx += (mouseX-width/2)/(2*step*sc*realm);
    ty += (mouseY-height/2)/(2*step*sc*realm);

    sc -= (step/30.0)*sc;


  }  

}

void setup_occupancy(){

  println("fetching occupancy data");
  String[] lines = loadStrings("http://adamsmith.as/blurry_maps/logfoo/occupancy.txt");
  noo = lines.length;

  println("parsing "+noo+" items");

  PGraphicsOpenGL pgl = (PGraphicsOpenGL) g;  // g may change
  GL gl = pgl.beginGL();  // always use the GL object returned by beginGL
  gl.setSwapInterval(1); //set vertical sync on 

  vbuf = BufferUtil.newFloatBuffer(noo * 16);
  for (int i = 0; i < noo; i++) {
    String guts = lines[i].substring(lines[i].indexOf('(') + 1, lines[i].indexOf(')'));
    String[] coords = guts.split(",");

    int x = Integer.parseInt(coords[0]);
    int y = Integer.parseInt(coords[1]);
    int z = Integer.parseInt(coords[2]);
    float sz = 1.0/(1<<z);
    float hz = (1<<z)/2.0;

    vbuf.put( (x-hz+0.0)*sz );
    vbuf.put( (y-hz+0.0)*sz );

    vbuf.put( (x-hz+0.0)*sz );
    vbuf.put( (y-hz+1.0)*sz );

    vbuf.put( (x-hz+0.0)*sz );
    vbuf.put( (y-hz+1.0)*sz );

    vbuf.put( (x-hz+1.0)*sz );
    vbuf.put( (y-hz+1.0)*sz );


    vbuf.put( (x-hz+1.0)*sz );
    vbuf.put( (y-hz+1.0)*sz );

    vbuf.put( (x-hz+1.0)*sz );
    vbuf.put( (y-hz+0.0)*sz );

    vbuf.put( (x-hz+1.0)*sz );
    vbuf.put( (y-hz+0.0)*sz );

    vbuf.put( (x-hz+0.0)*sz );
    vbuf.put( (y-hz+0.0)*sz );
  }
  vbuf.rewind();

  gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
  gl.glVertexPointer(2, GL.GL_FLOAT, 0, vbuf);

  pgl.endGL();

}

void setup_trace() {
  println("fetching trace data");
  String[] lines = loadStrings("http://adamsmith.as/blurry_maps/logfoo/trace.txt");
  not = lines.length;
  start = (int)random(not);
  trace = new int[not][3];

  println("parsing "+not+" items");

  for (int i = 0; i < not; i++) {
    String guts = lines[i].substring(lines[i].indexOf('(') + 1, lines[i].indexOf(')'));
    String[] coords = guts.split(",");

    int x = Integer.parseInt(coords[0]);
    int y = Integer.parseInt(coords[1]);
    int z = Integer.parseInt(coords[2]);

    trace[i][0] = x;
    trace[i][1] = y;
    trace[i][2] = z;
  }


}






















